#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <semaphore.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#define FILENAME "fichier"
#define ROOMCAPACITY 5
#define MAXNAMELEN 11

#define CHECK(op) { if ((op) == -1) { perror(#op); exit(1); } }
#define NCHECK(op) { if ((op) == NULL) { perror(#op); exit(1); } }

struct reunion
{
    char *name;
    char **members;
    size_t member_count;
};


sem_t *open_filelock(void)
{
    return sem_open("/" FILENAME ".lock", O_CREAT, 0777, 1);
}

struct reunion *create_reunion(void)
{
    struct reunion *reunion = malloc(sizeof(struct reunion));
    NCHECK(reunion);
    
    reunion->name = malloc(MAXNAMELEN * sizeof(char));
    NCHECK(reunion->name);
    
    reunion->members = malloc(ROOMCAPACITY * sizeof(char*));
    NCHECK(reunion->members);

    reunion->members[0] = malloc(ROOMCAPACITY * MAXNAMELEN * sizeof(char));
    NCHECK(reunion->members[0]);

    for (size_t i = 1; i < ROOMCAPACITY; i++)
        reunion->members[i] = reunion->members[0] + i*MAXNAMELEN;

    return reunion;
}

void free_reunion(struct reunion *reunion)
{
    free(reunion->name);
    free(reunion->members[0]);
    free(reunion->members);
    free(reunion);
}

struct reunion *read_reunion(FILE *file)
{
    if (file == NULL)
        return NULL;

    struct reunion *reunion = create_reunion();

    if (fgets(reunion->name, MAXNAMELEN, file) == NULL)
    {
        free_reunion(reunion);
        return NULL;
    }
    reunion->name[strlen(reunion->name) - 1] = '\0'; // remove last '\n'

    reunion->member_count = 0;
    while (fgets(reunion->members[reunion->member_count], MAXNAMELEN, file))
    {
        size_t len = strlen(reunion->members[reunion->member_count]);
        // remove last '\n'
        reunion->members[reunion->member_count][len-1] = '\0';
        reunion->member_count++;
    }

    reunion->member_count--; // last line is empty
    
    return reunion;
}

void write_reunion(FILE *file, struct reunion *reunion)
{
    fprintf(file, "%s\n", reunion->name);

    for (size_t i = 0; i < reunion->member_count; i++)
        fprintf(file, "%s\n", reunion->members[i]);
}

void control(void)
{
    sem_t *sem = open_filelock();


    int occupied = 2;

    while (true)
    {    
        int newoccupied;
        CHECK(sem_wait(sem));
        FILE *file = fopen(FILENAME, "r");
        if (file == NULL)
        {
            // switch (errno)
            // {
            // case EEXIST:
            //     occupied = false;
            //     break;
            
            // default:
            //     perror("opening " FILENAME);
            //     exit(EXIT_FAILURE);
            // }
            file = fopen(FILENAME, "w");
            newoccupied = false;
        }
        else
        {
            struct reunion *reunion = read_reunion(file);

            newoccupied = reunion != NULL && reunion->member_count > 0;

            if (reunion != NULL)
                free_reunion(reunion);
        }
        
        CHECK(fclose(file));

        CHECK(sem_post(sem));

        if (occupied == newoccupied)
            continue;

        occupied = newoccupied;
        if (occupied)
            printf("salle occupée\n");
        else
            printf("salle libre\n");
    }    
}

void cleanup(void)
{
    // no error check, just cleanup
    sem_unlink("/" FILENAME ".lock");
    CHECK(unlink(FILENAME));
}

void meet(char *meeting, char *name, int duration)
{
    if (strlen(meeting) >= MAXNAMELEN || strlen(name) >= MAXNAMELEN)
    {
        fprintf(stderr, "error: strings len must be < %d\n", MAXNAMELEN);
        exit(EXIT_FAILURE);
    }
    
    if (duration < 0)
    {
        fprintf(stderr, "error: %d < 0\n", duration);
        exit(EXIT_FAILURE);
    }

    sem_t *sem = open_filelock();

    FILE *file;
    struct reunion *reunion;

    // begin meeting
    while (true)
    {
        CHECK(sem_wait(sem));

        file = fopen(FILENAME, "r");
        // if (file == NULL && errno != EEXIST)
        // {
        //     perror("opening " FILENAME);
        //     exit(EXIT_FAILURE);
        // }

        reunion = read_reunion(file);
        if (reunion == NULL)
        {
            reunion = create_reunion();
            strncpy(reunion->name, meeting, MAXNAMELEN);
        }

        if (file != NULL)
            CHECK(fclose(file));

        if (!strcmp(reunion->name, meeting))
            break; // we found our meeting

        CHECK(sem_post(sem));
    }

    if (reunion->member_count >= ROOMCAPACITY)
    {
        fprintf(stderr, "error: full room\n");
        CHECK(sem_post(sem));
        exit(EXIT_FAILURE);
    }

    strncpy(reunion->members[reunion->member_count++], name, MAXNAMELEN);

    NCHECK(file = fopen(FILENAME, "w"));
    write_reunion(file, reunion);

    free_reunion(reunion);

    CHECK(fclose(file));
    CHECK(sem_post(sem));

    // wait the duration
    CHECK(usleep(duration));

    // end meeting
    CHECK(sem_wait(sem));
    NCHECK(file = fopen(FILENAME, "r"));

    reunion = read_reunion(file);
    if (reunion == NULL)
    {
        fprintf(stderr, "error: no meeting to remove '%s' from\n", name);
        exit(EXIT_FAILURE);
    }

    CHECK(fclose(file));

    size_t nameidx = 0;
    while (nameidx < ROOMCAPACITY && strcmp(reunion->members[nameidx], name))
        nameidx++;
    if (nameidx == ROOMCAPACITY)
    {
        fprintf(stderr, "error: '%s' is not anymore in the meeting\n", name);
        exit(EXIT_FAILURE);
    }

    memmove(reunion->members + nameidx, reunion->members + nameidx + 1,
            (ROOMCAPACITY - nameidx - 1)*sizeof(char*));
    
    NCHECK(file = fopen(FILENAME, "w"));
    write_reunion(file, reunion);
    
    CHECK(fclose(file));
    CHECK(sem_post(sem));
    CHECK(sem_close(sem));
}

int main(int argc, char *argv[])
{
    switch (argc)
    {
    case 2:
        if (!strcmp(argv[1], "controler"))
        {
            control();
        }
        else if (!strcmp(argv[1], "nettoyer"))
        {
            cleanup();
        }
        else
        {
            fprintf(stderr, "bad usage: %s controler|nettoyer\n", argv[0]);
            return EXIT_FAILURE;
        }
        break;

    case 4:
        meet(argv[1], argv[2], atoi(argv[3]) * 1000);
        break;
    
    default:
        fprintf(stderr, "bad usage: %s <réunion> <nom> <durée>\n"
                        "           %s controler|nettoyer\n",
                        argv[0], argv[0]);
        return EXIT_FAILURE;
    }
    
    return EXIT_SUCCESS;
}

